# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import json
import os
from os.path import exists
from os.path import join

from io import StringIO

from calmjs.toolchain import Spec
from calmjs.toolchain import CONFIG_JS_FILES
from calmjs.vlqsm import SourceWriter
from calmjs.npm import get_npm_version
from calmjs.utils import pretty_logging

from calmjs.rjs.ecma import parse
from calmjs.rjs import toolchain

from calmjs.testing import utils
from calmjs.testing import mocks


class SpecUpdateSourceMapTestCase(unittest.TestCase):
    """
    A function for updating the spec with a source map for target keys,
    in such a way that makes it compatible with the base system.
    """

    def test_spec_update_source_map_standard_modules_base(self):
        source_map = {
            'standard/module': 'standard/module',
            'standard.module': 'standard.module',
        }
        spec = {}

        toolchain.spec_update_source_map(spec, source_map, 'source_key')
        self.assertEqual(spec, {
            'source_key': {
                'standard/module': 'standard/module',
                'standard.module': 'standard.module',
            }
        })

    def test_spec_update_source_map_standard_modules_id(self):
        source_map = {
            'standard/module': 'standard/module',
        }
        base_map = {}
        spec = {'source_key': base_map}

        toolchain.spec_update_source_map(spec, source_map, 'source_key')
        self.assertIs(spec['source_key'], base_map)
        self.assertEqual(base_map, {
            'standard/module': 'standard/module',
        })

    def test_spec_update_source_map_plugin_modules(self):
        source_map = {
            'plugin/module!argument': 'some/filesystem/path',
            'text!argument': 'some/text/file.txt',
        }
        spec = {}

        toolchain.spec_update_source_map(spec, source_map, 'source_key')
        self.maxDiff = 123123
        self.assertEqual(spec, {
            'requirejs_plugins': {
                'plugin/module': {
                    'plugin/module!argument': 'some/filesystem/path',
                },
                'text': {
                    'text!argument': 'some/text/file.txt',
                },
            },
            'source_key': {
            },
        })

        toolchain.spec_update_source_map(spec, {
            'text!argument2': 'some/text/file2.txt',
        }, 'source_key')
        self.assertEqual(spec['requirejs_plugins']['text'], {
            'text!argument': 'some/text/file.txt',
            'text!argument2': 'some/text/file2.txt',
        })


class ToolchainBootstrapTestCase(unittest.TestCase):
    """
    Test the bootstrap function
    """

    def test_runtime_name(self):
        # seems redundant, but...
        platform = 'posix'
        self.assertEqual(toolchain.get_rjs_runtime_name(platform), 'r.js')
        platform = 'win32'
        self.assertEqual(toolchain.get_rjs_runtime_name(platform), 'r.js.cmd')


class TranspilerTestCase(unittest.TestCase):
    """
    Test various "transpilers"  in the toolchain module
    """

    def test_transpile_generic_to_umd_node_amd_compat_rjs_basic(self):
        source = StringIO(
            'var dummy = function () {};\n'
            '\n'
            'exports.dummy = dummy;\n'
        )
        target = SourceWriter(StringIO())
        spec = Spec()

        toolchain._transpile_generic_to_umd_node_amd_compat_rjs(
            spec, source, target)

        self.assertEqual(target.getvalue().splitlines()[2:8], [
            '        var exports = {};',
            '        var dummy = function () {};',
            '',
            '        exports.dummy = dummy;',
            '',
            '        return exports;',
        ])

    def test_transpile_generic_to_umd_node_amd_compat_rjs_strict(self):
        source = StringIO(
            '"use strict";\n'
            'var dummy = function () {};\n'
            '\n'
            'exports.dummy = dummy;\n'
        )
        target = SourceWriter(StringIO())
        spec = Spec()

        toolchain._transpile_generic_to_umd_node_amd_compat_rjs(
            spec, source, target)

        self.assertEqual(target.getvalue().splitlines()[2:9], [
            '        "use strict";',
            '        var exports = {};',
            '        var dummy = function () {};',
            '',
            '        exports.dummy = dummy;',
            '',
            '        return exports;',
        ])

    def test_transpile_generic_to_umd_node_amd_compat_rjs_strict_padded(self):
        """
        Show that padding empty lines will simply be used as padding,
        and that the 'use strict' declaration will just be moved down.
        This is to provide an option for developers so that line numbers
        of the "transpiled" code can be mapped directly back to its
        original.

        Naturally, spec will also need to have no indent specified.
        """

        original = (
            'var dummy = function () {};\n'
            'exports.dummy = dummy;'
        )

        source = StringIO(
            '"use strict";\n'
            '\n'
            '\n'
            '\n' +
            original + '\n'
        )
        target = SourceWriter(StringIO())
        spec = Spec(transpile_no_indent=True)

        toolchain._transpile_generic_to_umd_node_amd_compat_rjs(
            spec, source, target)

        self.assertEqual(
            '\n'.join(target.getvalue().splitlines()[4:6]), original)

    def test_transpile_generic_to_umd_node_amd_compat_rjs_basic_padded(self):
        """
        Similar as above, but insufficient lines will just result in
        not much happening...
        """

        source = StringIO(
            '\n'
            'var dummy = function () {};\n'
            '\n'
            'exports.dummy = dummy;\n'
        )
        target = SourceWriter(StringIO())
        spec = Spec()

        toolchain._transpile_generic_to_umd_node_amd_compat_rjs(
            spec, source, target)

        self.assertEqual(target.getvalue().splitlines()[2:8], [
            '        var exports = {};',  # no break after this.
            '        var dummy = function () {};',
            '',
            '        exports.dummy = dummy;',
            '',
            '        return exports;',
        ])

        self.assertEqual(target.mappings[:7], [
            [], [], [], [(8, 0, 1, 0)], [(0, 0, 1, 0)], [(8, 0, 1, 0)], []])

        target_main = SourceWriter(StringIO())
        toolchain._rjs_transpiler(spec, source, target_main)
        self.assertEqual(target.getvalue(), target_main.getvalue())

    def test_transpile_skip_on_amd_newline(self):
        source = StringIO(
            "\n"
            "define(['jquery'], function($) {\n"
            "    'use strict';\n"
            "    return {'testing': {}}\n"
            "});\n"
        )
        target = StringIO()
        spec = Spec()
        toolchain._rjs_transpiler(spec, source, target)
        self.assertEqual(source.getvalue(), target.getvalue())

    def test_transpile_skip_on_amd_strict_top(self):
        source = StringIO(
            "'use strict';\n"
            "define(['jquery'], function($) {\n"
            "    return {'testing': {}}\n"
            "});\n"
        )
        target = StringIO()
        spec = Spec()
        toolchain._rjs_transpiler(spec, source, target)
        self.assertEqual(source.getvalue(), target.getvalue())

    def test_modname_source_target_to_modpath(self):
        rjs = toolchain.RJSToolchain()
        spec = Spec()
        self.assertEqual(rjs.modname_source_target_to_modpath(
            spec,
            'example/module',
            '/tmp/src/example/module', '/tmp/build/example/module'),
            'example/module'
        )
        self.assertEqual(rjs.modname_source_target_to_modpath(
            spec,
            'example/module',
            'empty:', '/tmp/build/example/module'),
            'empty:'
        )

    def test_transpile_modname_source_target_normal(self):
        modname = 'module'
        src_file = join(utils.mkdtemp(self), 'module.js')
        tgt_file = 'module.js'
        tgt_dir = utils.mkdtemp(self)
        spec = Spec(build_dir=tgt_dir)

        with open(src_file, 'w') as fd:
            fd.write('console.log("Hello");')

        rjs = toolchain.RJSToolchain()
        rjs.transpile_modname_source_target(spec, modname, src_file, tgt_file)

        with open(join(tgt_dir, tgt_file)) as fd:
            # The transpiler will mutate it.
            result = fd.read()

        self.assertNotEqual('console.log("Hello");', result)
        self.assertIn('console.log("Hello");', result)

    def test_transpile_modname_source_target_empty(self):
        modname = 'module'
        src_file = 'empty:'
        tgt_dir = utils.mkdtemp(self)
        tgt_file = join(tgt_dir, 'module.js')
        spec = Spec(build_dir=tgt_dir)
        rjs = toolchain.RJSToolchain()
        rjs.transpile_modname_source_target(spec, modname, src_file, tgt_file)
        self.assertFalse(exists(tgt_file))


class ToolchainCompilePluginTestCase(unittest.TestCase):
    """
    Test the compile_plugin method
    """

    def test_toolchain_instance_transpile_skip_on_amd(self):
        source = StringIO(
            "define(['jquery'], function($) {\n"
            "    'use strict';\n"
            "    return {'testing': {}}\n"
            "});\n"
        )
        target = StringIO()
        spec = Spec()
        rjs = toolchain.RJSToolchain()
        rjs.transpiler(spec, source, target)
        self.assertEqual(source.getvalue(), target.getvalue())

    def test_compile_plugin_base(self):
        build_dir = utils.mkdtemp(self)
        src_dir = utils.mkdtemp(self)
        src = join(src_dir, 'mod.js')

        with open(src, 'w') as fd:
            fd.write('hello world')

        # prepare targets
        target1 = 'mod1.txt'
        target2 = join('namespace', 'mod2.txt')
        target3 = join('nested', 'namespace', 'mod3.txt')
        target4 = 'namespace.mod4.txt'

        rjs = toolchain.RJSToolchain()
        spec = {
            'build_dir': build_dir,
            toolchain.RJS_LOADER_PLUGIN_REGISTRY: rjs.loader_plugin_registry,
        }
        rjs.compile_plugin(spec, [
            ('text!mod1.txt', src, target1, 'mod1'),
            ('text!namespace/mod2.txt', src, target2, 'mod2'),
            ('text!nested/namespace/mod3.txt', src, target3, 'mod3'),
            ('text!namespace.mod4.txt', src, target4, 'mod4'),
        ])

        self.assertTrue(exists(join(build_dir, target1)))
        self.assertTrue(exists(join(build_dir, target2)))
        self.assertTrue(exists(join(build_dir, target3)))
        self.assertTrue(exists(join(build_dir, target4)))

    def test_compile_plugin_error(self):
        build_dir = utils.mkdtemp(self)
        src_dir = utils.mkdtemp(self)
        src = join(src_dir, 'mod.js')

        with open(src, 'w') as fd:
            fd.write('hello world')

        # prepare targets
        target = 'target.txt'

        rjs = toolchain.RJSToolchain()
        spec = {
            'build_dir': build_dir,
            toolchain.RJS_LOADER_PLUGIN_REGISTRY: rjs.loader_plugin_registry,
        }
        with self.assertRaises(TypeError):
            # This normally shouldn't happen, and for now the method
            # will not trap exceptions.
            rjs.compile_plugin(spec, [
                ('unregistered/mod!target.txt', src, target, 'target.txt'),
            ])


class ToolchainBaseUnitTestCase(unittest.TestCase):
    """
    Test the base functions in the toolchain.
    """

    def test_modname_source_target_to_modpath(self):
        modname = 'module'
        source = 'somemodule'
        rjs = toolchain.RJSToolchain()
        spec = Spec()
        self.assertEqual(
            rjs.modname_source_target_to_modpath(spec, modname, source, ''),
            'module',
        )

    def test_modname_source_target_to_modpath_empty(self):
        modname = 'module'
        source = 'empty:'
        rjs = toolchain.RJSToolchain()
        spec = Spec()
        self.assertEqual(
            rjs.modname_source_target_to_modpath(spec, modname, source, ''),
            'empty:',
        )


@unittest.skipIf(get_npm_version() is None, "npm is unavailable")
class ToolchainUnitTestCase(unittest.TestCase):
    """
    Just testing out the toolchain units.
    """

    def test_prepare_failure_manual(self):
        rjs = toolchain.RJSToolchain()
        spec = Spec(rjs_bin='/no/such/path')
        with self.assertRaises(RuntimeError) as e:
            rjs.prepare(spec)

        self.assertEqual(
            str(e.exception),
            "'/no/such/path' does not exist; cannot be used as '%s' binary" % (
                rjs.rjs_bin
            ),
        )

    def test_prepare_failure_which_fail(self):
        utils.stub_os_environ(self)
        utils.remember_cwd(self)

        # must go to a directory where r.js is guaranteed to not be
        # available through node_modules or the environmental PATH
        os.environ['NODE_PATH'] = ''
        os.environ['PATH'] = ''
        tmpdir = utils.mkdtemp(self)
        os.chdir(tmpdir)

        rjs = toolchain.RJSToolchain()
        spec = Spec()
        with self.assertRaises(RuntimeError) as e:
            rjs.prepare(spec)

        self.assertEqual(str(e.exception), "unable to locate '%s'" % (
            rjs.rjs_bin
        ))

    def test_prepare_failure_export_target(self):
        tmpdir = utils.mkdtemp(self)
        rjs = toolchain.RJSToolchain()

        with open(join(tmpdir, 'r.js'), 'w'):
            # mock a r.js file.
            pass

        spec = Spec(build_dir=tmpdir)
        spec[rjs.rjs_bin_key] = join(tmpdir, 'r.js')

        with self.assertRaises(RuntimeError) as e:
            rjs.prepare(spec)
        self.assertEqual(
            str(e.exception), "'export_target' not found in spec")

        # what can possibly go wrong?
        spec['export_target'] = join(spec[rjs.rjs_bin_key], 'build.js')

        with self.assertRaises(RuntimeError) as e:
            rjs.prepare(spec)
        self.assertEqual(
            str(e.exception), "'export_target' will not be writable")

        spec['export_target'] = join(tmpdir, 'build.js')

        with self.assertRaises(RuntimeError) as e:
            rjs.prepare(spec)
        self.assertEqual(
            str(e.exception), "'export_target' must not be same as "
            "'build_manifest_path'")

    def test_assemble_null(self):
        tmpdir = utils.mkdtemp(self)

        with open(join(tmpdir, 'r.js'), 'w'):
            # mock a r.js file.
            pass

        spec = Spec(
            # this is not written
            export_target=join(tmpdir, 'bundle.js'),
            build_dir=tmpdir,
            transpiled_modpaths={},
            bundled_modpaths={},
            plugins_modpaths={},
            transpiled_targets={},
            bundled_targets={},
            plugins_targets={},
            export_module_names=[],
        )

        rjs = toolchain.RJSToolchain()
        spec[rjs.rjs_bin_key] = join(tmpdir, 'r.js')
        rjs.prepare(spec)
        rjs.assemble(spec)

        self.assertTrue(exists(join(tmpdir, 'build.js')))
        self.assertTrue(exists(join(tmpdir, 'config.js')))
        self.assertEqual(spec[CONFIG_JS_FILES], [join(tmpdir, 'config.js')])

        with open(join(tmpdir, 'build.js')) as fd:
            # strip off the header and footer
            build_js = json.loads(''.join(fd.readlines()[1:-1]))

        with open(join(tmpdir, 'config.js')) as fd:
            # strip off the header and footer
            config_js = json.loads(''.join(fd.readlines()[4:-10]))

        self.assertEqual(build_js['paths'], {})
        self.assertEqual(build_js['include'], [])

        self.assertEqual(config_js['paths'], {})
        self.assertEqual(config_js['include'], [])

    def test_prepare_assemble(self):
        tmpdir = utils.mkdtemp(self)

        with open(join(tmpdir, 'r.js'), 'w'):
            # mock a r.js file.
            pass

        spec = Spec(
            # this is not written
            export_target=join(tmpdir, 'bundle.js'),
            build_dir=tmpdir,
            transpiled_modpaths={
                'example/module': '/path/to/src/example/module'
            },
            bundled_modpaths={
                'bundled_pkg': '/path/to/bundled/index',
                'bundled_empty': 'empty:',
            },
            plugins_modpaths={
                'loader/plugin!resource/name': '/resource/name'
            },
            transpiled_targets={
                'example/module': '/path/to/src/example/module.js',
            },
            bundled_targets={
                'bundled_pkg': '/path/to/bundled/index.js',
                'bundled_txt': '/path/to/bundled/txt',
                'bundled_dir': '/path/to/bundled/dir.js',
                'bundled_empty': 'empty:',
            },
            plugins_targets={
                'resource/name': '/resource/name',
            },
            export_module_names=[
                'example/module',
                'bundled_dir',
                'bundled_pkg',
                'bundled_txt',
                'bundled_empty',
                'loader/plugin!resource/name',
            ],
        )

        # we are going to fake the is_file checks
        utils.stub_item_attr_value(
            self, toolchain, 'isfile', lambda x: not x.endswith('dir.js'))

        rjs = toolchain.RJSToolchain()
        spec[rjs.rjs_bin_key] = join(tmpdir, 'r.js')
        rjs.prepare(spec)

        # skip the compile step as those entries are manually applied.
        with pretty_logging(logger='calmjs.rjs', stream=mocks.StringIO()) as s:
            # the parser will try to load the file
            rjs.assemble(spec)

        self.assertIn('No such file or directory', s.getvalue())
        self.assertIn(
            join(*('path/to/src/example/module.js'.split('/'))),
            s.getvalue(),
        )

        self.assertTrue(exists(join(tmpdir, 'build.js')))
        self.assertTrue(exists(join(tmpdir, 'config.js')))

        with open(join(tmpdir, 'build.js')) as fd:
            # strip off the header and footer as this is for r.js
            build_js = json.loads(''.join(fd.readlines()[1:-1]))

        with open(join(tmpdir, 'config.js')) as fd:
            # strip off the header and footer as this is for r.js
            config_js = json.loads(''.join(fd.readlines()[4:-10]))

        self.assertEqual(build_js['paths'], {
            'bundled_empty': 'empty:',
        })
        self.assertEqual(build_js['include'], [
            'example/module',
            'bundled_dir',
            'bundled_pkg',
            'bundled_txt',
            'bundled_empty',
            'loader/plugin!resource/name',
        ])

        self.assertEqual(config_js['paths'], {
            'example/module': '/path/to/src/example/module.js?',
            'bundled_pkg': '/path/to/bundled/index.js?',
            'bundled_txt': '/path/to/bundled/txt',
            'bundled_dir': '/path/to/bundled/dir.js',
            'resource/name': '/resource/name',
        })
        self.assertEqual(config_js['include'], [])

    def test_prepare_rjs_plugin_key(self):
        tmpdir = utils.mkdtemp(self)
        rjs = toolchain.RJSToolchain()

        with open(join(tmpdir, 'r.js'), 'w'):
            # mock a r.js file.
            pass

        spec = Spec(
            # this is not written
            export_target=join(tmpdir, 'bundle.js'),
            build_dir=tmpdir,
            transpiled_modpaths={},
            bundled_modpaths={},
            export_module_names=[],
        )
        spec[rjs.rjs_bin_key] = join(tmpdir, 'r.js')
        spec[toolchain.REQUIREJS_PLUGINS] = {
            'text': {
                'text!namespace/module/path.txt': '/namespace/module/path.txt',
            },
            'unsupported/unknown_plugin': {
                'also this is an invalid value': '/some/path',
            },
        }

        with pretty_logging(logger='calmjs.rjs', stream=mocks.StringIO()) as s:
            rjs.prepare(spec)

        self.assertEqual(spec['plugin_source_map'], {
            'text!namespace/module/path.txt': '/namespace/module/path.txt',
        })

        logs = s.getvalue()
        self.assertIn("DEBUG", logs)
        self.assertIn("found handler for 'text' loader plugin", logs)
        self.assertIn("WARNING", logs)
        self.assertIn(
            "handler for 'unsupported/unknown_plugin' loader plugin not found",
            logs)
        self.assertIn("also this is an invalid value", logs)

    def assemble_spec_config(self, **kw):
        # for the assemble related tests.
        tmpdir = utils.mkdtemp(self)
        build_dir = utils.mkdtemp(self)
        rjs = toolchain.RJSToolchain()

        export_target = join(build_dir, 'export.js')
        build_manifest_path = join(build_dir, 'build.js')
        node_config_js = join(build_dir, 'node_config.js')
        requirejs_config_js = join(build_dir, 'requirejs_config.js')

        with open(join(tmpdir, 'r.js'), 'w'):
            pass

        with open(join(build_dir, 'module1.js'), 'w') as fd:
            fd.write(
                "define(['jquery', 'underscore', 'some.pylike.module'], "
                "function(jquery, underscore, module) {"
                "});"
            )

        with open(join(build_dir, 'module2.js'), 'w') as fd:
            fd.write(
                "define(['module1', 'underscore'], "
                "function(module1, underscore) {"
                "});"
            )

        with open(join(build_dir, 'module3.js'), 'w') as fd:
            fd.write(
                "'use strict';\n"
                "var $ = require('jquery');\n"
                "var _ = require('underscore');\n"
                "var module2 = require('module2');\n"
            )

        spec = Spec(
            build_dir=build_dir,
            export_target=export_target,
            build_manifest_path=build_manifest_path,
            node_config_js=node_config_js,
            requirejs_config_js=requirejs_config_js,
            transpiled_modpaths={
                'module1': 'module1',
                'module2': 'module2',
                'module3': 'module3',
            },
            # these are not actually transpiled sources, but will fit
            # with the purposes of this test.
            transpiled_targets={
                'module1': 'module1.js',
                'module2': 'module2.js',
                'module3': 'module3.js',
            },
            # the "bundled" names were specified to be omitted.
            bundled_modpaths={},
            bundled_targets={},
            plugins_modpaths={},
            plugins_targets={},
            export_module_names=['module1', 'module2', 'module3'],
            **kw
        )
        spec[rjs.rjs_bin_key] = join(tmpdir, 'r.js')
        rjs.assemble(spec)

        # the main config file
        # check that they all exists
        self.assertTrue(exists(build_manifest_path))
        self.assertTrue(exists(node_config_js))
        self.assertTrue(exists(requirejs_config_js))

        # only checking the build_manifest version, as the node config
        # version is not that much different.
        with open(build_manifest_path) as fd:
            build_tree = parse(fd.read())

        # this is the node for the json in the build file
        build_js = json.loads(build_tree.children()[0].children()[0].to_ecma())

        with open(requirejs_config_js) as fd:
            config_tree = parse(fd.read())

        # this is the node for json in the config file
        config_js = json.loads(
            config_tree.children()[0].children()[0].children()[0].children(
                )[2].children()[0].children()[1].to_ecma())

        return build_js, config_js

    def test_assemble_standard(self):
        with pretty_logging(logger='calmjs.rjs', stream=mocks.StringIO()) as s:
            build_js, config_js = self.assemble_spec_config()

        self.assertIn('ERROR', s.getvalue())
        self.assertIn(
            "source file(s) referenced modules that are missing in the "
            "build directory: 'jquery', 'some.pylike.module', 'underscore'",
            s.getvalue()
        )

        self.assertEqual(build_js['paths'], {})
        self.assertEqual(config_js['paths'], {
            'module1': 'module1.js?',
            'module2': 'module2.js?',
            'module3': 'module3.js?',
        })

    def test_assemble_standard_emptied(self):
        with pretty_logging(logger='calmjs.rjs', stream=mocks.StringIO()) as s:
            build_js, config_js = self.assemble_spec_config(
                stub_missing_with_empty=1
            )

        self.assertNotIn('ERROR', s.getvalue())
        self.assertIn(
            "source file(s) referenced modules that are missing in the "
            "build directory: 'jquery', 'some.pylike.module', 'underscore'",
            s.getvalue()
        )

        self.assertEqual(build_js['paths'], {
            'jquery': 'empty:',
            'some.pylike.module': 'empty:',
            'underscore': 'empty:',
        })
        self.assertEqual(config_js['paths'], {
            'module1': 'module1.js?',
            'module2': 'module2.js?',
            'module3': 'module3.js?',
            'jquery': 'empty:',
            'some.pylike.module': 'empty:',
            'underscore': 'empty:',
        })



# -*- coding: utf-8 -*-
import unittest
from os import mkdir
from os import makedirs
from os.path import exists
from os.path import join

from calmjs.rjs import plugin
from calmjs.utils import pretty_logging

from calmjs.testing.utils import mkdtemp
from calmjs.testing.mocks import StringIO


class LoaderPluginHandlerTestCase(unittest.TestCase):

    def test_config_paths(self):
        handler = plugin.LoaderPluginHandler(None)
        self.assertEqual(handler.modname_target_to_config_paths(
            'example/path', 'example/path.js'),
            {'example/path': 'example/path.js?'},
        )
        self.assertEqual(handler.modname_source_to_config_paths(
            'example/path', 'example/path.js'),
            {'example/path': 'example/path.js?'},
        )

    def test_others(self):
        handler = plugin.LoaderPluginHandler(None)
        modname_modpath = ('example/path', 'example/path')
        self.assertEqual(
            handler.modname_modpath_to_config_paths(*modname_modpath),
            dict([modname_modpath]),
        )
        with self.assertRaises(NotImplementedError):
            handler(None, None, None, None, None, None)


class TextLoaderPluginTestCase(unittest.TestCase):

    def test_strip_plugin(self):
        f = plugin.TextPlugin(None).strip_plugin
        self.assertEqual(f('file.txt'), 'file.txt')
        self.assertEqual(f('text!file.txt'), 'file.txt')
        self.assertEqual(f('text!file.txt!strip'), 'file.txt')

    def test_strip_plugin_unstripped_values(self):
        f = plugin.TextPlugin(None).strip_plugin
        self.assertEqual(f('/file.txt'), '/file.txt')
        self.assertEqual(f('/text!file.txt'), '/text!file.txt')
        self.assertEqual(f('/text!file.txt!strip'), '/text!file.txt!strip')

    def test_strip_plugin_empty(self):
        f = plugin.TextPlugin(None).strip_plugin
        # this should be invalid, but we are forgiving
        self.assertEqual(f(''), '')
        self.assertEqual(f('text!'), '')

    def test_requirejs_text_issue123_handling(self):
        f = plugin.TextPlugin(None).requirejs_text_issue123
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(f('file'), ['file', ''])
            self.assertEqual(f('dir/text'), ['dir/text', ''])
            self.assertEqual(f('dir/text.txt'), ['dir/text', 'txt'])
            self.assertEqual(f('dotted.dir/text'), ['dotted', 'dir/text'])
            self.assertEqual(
                f('dotted.dir/text.txt'), ['dotted.dir/text', 'txt'])

        self.assertEqual(stream.getvalue(), '')

        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            # the following also implement the stripping of trailing
            # dots which requirejs-text doesn't support correctly,
            # and with requirejs>=2.0.13 will fail consistently.
            self.assertEqual(f('file.'), ['file', ''])
            self.assertEqual(f('dir/text.'), ['dir/text', ''])
            self.assertEqual(f('dotted.dir/text.'), ['dotted.dir/text', ''])

        # ensure the complaining loudly is done.
        self.assertIn('WARNING', stream.getvalue())
        self.assertIn('trailing', stream.getvalue())

    def test_empty_modname_target_to_config_paths(self):
        f = plugin.TextPlugin(None).modname_target_to_config_paths
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(
                f('text!', 'text!'),
                {'': ''},  # we are following our own rules...
            )
        self.assertEqual(stream.getvalue(), '')

    def test_modname_target_to_config_paths(self):
        f = plugin.TextPlugin(None).modname_target_to_config_paths
        self.modname_x_to_config_paths(f)

    def test_modname_source_to_config_paths(self):
        f = plugin.TextPlugin(None).modname_source_to_config_paths
        self.modname_x_to_config_paths(f)

    def modname_x_to_config_paths(self, f):
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(
                f('text!file', 'text!file'),
                {'file': 'file'},
            )
            self.assertEqual(
                f('text!file.txt', 'text!file.txt'),
                {
                    'file': 'file',
                    'file.txt': 'file.txt',
                },
            )
            self.assertEqual(
                f('text!dir/file.txt', 'text!dir/file.txt'),
                {
                    'dir/file': 'dir/file',
                    'dir/file.txt': 'dir/file.txt',
                },
            )
            self.assertEqual(
                f('text!dir/file.txt', '/some/path/dir/file.txt'),
                {
                    'dir/file': '/some/path/dir/file',
                    'dir/file.txt': '/some/path/dir/file.txt',
                },
            )
            self.assertEqual(
                f('text!dir/file', '/some/path/dir/file'),
                {'dir/file': '/some/path/dir/file'},
            )
            self.assertEqual(
                f('text!dir/file', '/some/path/dir/file!strip'),
                {'dir/file': '/some/path/dir/file!strip'},
            )

        self.assertEqual(stream.getvalue(), '')

    def test_modname_target_to_config_paths_mismatch_file_ext(self):
        # this will blow up, actually, since requesting for file.htm
        # will NOT resolve it to file.html - no configuration can be
        # currently done to produce a working mapping.
        f = plugin.TextPlugin(None).modname_target_to_config_paths
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(
                f('text!file.htm', 'text!some.dotted/dir/file.html'),
                {'file.htm': 'some.dotted/dir/file.html'},
            )
        err = stream.getvalue()
        self.assertIn('WARNING', err)
        self.assertIn('no possible workaround', err)

    def test_modname_target_to_config_paths_mismatch_dir_ext_file_noext(self):
        f = plugin.TextPlugin(None).modname_target_to_config_paths
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(
                {
                    'file': 'some.dotted/dir/file',
                    'file.ns': 'some.dotted/dir/file.ns',
                    'file.ns/html': 'some.dotted/dir/file.ns/html',
                },
                f('text!file.ns/html', 'text!some.dotted/dir/file.ns/html'),
            )
        # this one provides both the dot stripped and the underlying
        # directory, to cover both case (unless this screws up some
        # other thing).
        err = stream.getvalue()
        # though since this should work, better to warn about this.
        self.assertIn('WARNING', err)
        self.assertIn(
            "warning triggered by mapping config.paths from "
            "modpath 'text!file.ns/html' to "
            "target 'text!some.dotted/dir/file.ns/html'", err
        )
        self.assertIn("unsupported values provided", err)
        self.assertIn("potentially working mitigations applied", err)
        self.assertIn('text!file.ns/html', err)
        self.assertIn('text!some.dotted/dir/file.ns/html', err)

    def test_modname_target_to_config_paths_warning(self):
        f = plugin.TextPlugin(None).modname_target_to_config_paths
        # There are cases where if the python style namespace separator
        # is used for the generated path, and the final fragment has no
        # further '.' characters, WILL result in a complete mismatch of
        # the key to the expected target.  That said, this particular
        # type of mapping doesn't seem to be supported at all by the
        # requirejs-text plugin.
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(
                f('text!some.target/file', '/src/some/target/file'),
                {'some.target/file': '/src/some/target/file'},
                # Do nothing as this type of conversion/mapping doesn't
                # work no matter what, because requirejs and/or the text
                # loader plugin fails at tracking directories if it has
                # a '.' somewhere, so the loader gets confused and just
                # don't see the mapping entry.

                # To mitigate, ensure the provided files or paths
                # provided have a filename extension and does not end
                # with a ``.`` character.
            )

        # ensure that is logged
        err = stream.getvalue()
        self.assertIn('WARNING', err)
        self.assertIn('no possible workaround', err)

    def test_modname_target_to_config_paths_info_no_false_positive(self):
        # like the htm -> html example
        f = plugin.TextPlugin(None).modname_target_to_config_paths
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(
                f('text!some/file.html', '/src/some/target/file.rst'),
                {'some/file.html': '/src/some/target/file.rst'},
            )
        err = stream.getvalue()
        self.assertIn('WARNING', err)
        self.assertIn('no possible workaround', err)

    def test_modname_target_to_config_paths_further_insanity(self):
        # Again, the mapping just do not work.  Mapping from 'some' to
        # '/src/some' will not suddenly make 'contents/html' map to
        # 'contents/rst'.
        f = plugin.TextPlugin(None).modname_target_to_config_paths
        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            self.assertEqual(
                {'some.contents/html': '/src/some.contents/rst'},
                f('text!some.contents/html', '/src/some.contents/rst'),
            )
        err = stream.getvalue()
        self.assertIn('WARNING', err)
        self.assertIn('no possible workaround', err)

    def test_basic(self):
        # the target is 'text_file.txt'
        build_dir = mkdtemp(self)
        srcdir = mkdtemp(self)
        spec = {'build_dir': build_dir}
        source = join(srcdir, 'text_file.txt')

        with open(source, 'w') as fd:
            fd.write('a text file\n')

        toolchain = None  # this one is not necessary for text.
        modname = 'text!text_file.txt'
        target = 'text!text_file.txt'
        modpath = 'text!text_file.txt'

        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            result = plugin.TextPlugin(None)(
                toolchain, spec, modname, source, target, modpath)
        self.assertEqual(stream.getvalue(), '')

        self.assertTrue(exists(join(build_dir, 'text_file.txt')))
        bundled_modpaths, bundled_targets, module_name = result

        self.assertEqual(bundled_modpaths, {
            'text!text_file.txt': 'text!text_file.txt',
        })
        self.assertEqual(bundled_targets, {
            'text_file': 'text_file',
            'text_file.txt': 'text_file.txt',
        })
        self.assertEqual(module_name, ['text!text_file.txt'])

    def test_nested(self):
        # the target is 'namespace/text_file.txt'
        build_dir = mkdtemp(self)
        srcdir = join(mkdtemp(self), 'namespace')
        mkdir(srcdir)
        spec = {'build_dir': build_dir}
        source = join(srcdir, 'text_file.txt')

        with open(source, 'w') as fd:
            fd.write('a text file\n')

        toolchain = None  # this one is not necessary for text.
        modname = 'text!namespace/text_file.txt'
        target = 'text!namespace/text_file.txt'
        modpath = 'text!namespace/text_file.txt'

        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            result = plugin.TextPlugin(None)(
                toolchain, spec, modname, source, target, modpath)
        self.assertEqual(stream.getvalue(), '')

        self.assertTrue(exists(join(build_dir, 'namespace', 'text_file.txt')))
        bundled_modpaths, bundled_targets, module_name = result
        self.assertEqual(bundled_modpaths, {
            'text!namespace/text_file.txt': 'text!namespace/text_file.txt',
        })
        self.assertEqual(bundled_targets, {
            'namespace/text_file': 'namespace/text_file',
            'namespace/text_file.txt': 'namespace/text_file.txt',
        })
        self.assertEqual(module_name, ['text!namespace/text_file.txt'])

    def test_dotted_namespace(self):
        # the target is 'dotted.ns/data'
        build_dir = mkdtemp(self)
        srcdir = join(mkdtemp(self), 'dotted.ns')
        mkdir(srcdir)
        spec = {'build_dir': build_dir}
        source = join(srcdir, 'data')

        with open(source, 'w') as fd:
            fd.write('a text file\n')

        toolchain = None  # this one is not necessary for text.
        modname = 'text!dotted.ns/data'
        target = 'text!dotted.ns/data'
        modpath = 'text!dotted.ns/data'

        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            result = plugin.TextPlugin(None)(
                toolchain, spec, modname, source, target, modpath)
        err = stream.getvalue()
        self.assertIn('WARNING', err)
        self.assertIn('potentially working mitigations applied', err)

        self.assertTrue(exists(join(build_dir, 'dotted.ns', 'data')))
        bundled_modpaths, bundled_targets, module_name = result
        self.assertEqual(bundled_modpaths, {
            'text!dotted.ns/data': 'text!dotted.ns/data',
        })
        # Yes, this is how bad requirejs-text is at figuring out what a
        # dot means within a path, how it misunderstand directories and
        # filename extensions within a path value.  Please refer to the
        # source code to note the issue at hand which is encapsulate by
        # text.requirejs_text_issue123 method.
        self.assertEqual(bundled_targets, {
            'dotted': 'dotted',
            'dotted.ns': 'dotted.ns',
            'dotted.ns/data': 'dotted.ns/data',
        })
        # Oh yeah, to confirm again that this is the expected value so
        # that the underlying loader will be able to load the target
        # dotted.ns/data.  This is especially critical for generation of
        # configuration for in-place testing/test files; see an earlier
        # test to show what kind of data other calmjs registries will
        # be able to produce.
        self.assertEqual(module_name, ['text!dotted.ns/data'])

    def test_dotted_namespace_mismatched_warning(self):
        # from src/dotted/ns/data to dotted.ns/data and this will blow
        # up completely (also not supported because requirejs-text issue
        # #123.
        build_dir = mkdtemp(self)
        srcdir = join(mkdtemp(self), 'dotted', 'ns')
        makedirs(srcdir)
        spec = {'build_dir': build_dir}
        source = join(srcdir, 'data')

        with open(source, 'w') as fd:
            fd.write('a text file\n')

        toolchain = None  # this one is not necessary for text.
        modname = 'text!dotted.ns/data'
        # note, this mismatch target and modpath shouldn't normally be
        # produced, however as the build script will have an empty paths
        # mapping, this wouldn't have worked as part of the build anyway.
        # So for plugin authors, do ensure the targets in the build
        # directory map directly to the intended locations.
        target = join('dotted', 'ns', 'data')
        modpath = 'text!dotted.ns/data'

        with pretty_logging('calmjs.rjs.plugin', stream=StringIO()) as stream:
            result = plugin.TextPlugin(None)(
                toolchain, spec, modname, source, target, modpath)
        err = stream.getvalue()
        self.assertIn('WARNING', err)
        self.assertIn("modpath 'text!dotted.ns/data'", err)
        # trimming off the ends because they are identical but the
        # actual values depend on os.sep.
        self.assertIn("target 'dotted", err)

        self.assertTrue(exists(join(build_dir, 'dotted', 'ns', 'data')))
        bundled_modpaths, bundled_targets, module_name = result
        self.assertEqual(bundled_modpaths, {
            'text!dotted.ns/data': 'text!dotted.ns/data',
        })
        # nothing was done, no mitigation applied as the generation of
        # a working config.paths is impossible
        self.assertEqual(bundled_targets, {
            'dotted.ns/data': target,  # 'dotted/ns/data'
        })
        self.assertEqual(module_name, ['text!dotted.ns/data'])
